import argparse
import numpy as np
import pylab as pl
import sys

from colorsys import hls_to_rgb
from pandas import DataFrame, read_csv
from scipy.stats import mannwhitneyu


class MeanAffectiveProfile(object):
    def __init__(self, iaps):
        self.iaps = iaps

    def get_iap_distribution(self):
        result = []

        for i in xrange(len(self.iaps[0]['normalised'])):
            result.append([iap['normalised'][i] for iap in self.iaps])

        return result

    def get_mean_profile(self):
        xn = reduce(
            lambda result, x: result + x['iai'],
            self.iaps,
            0
        ) / len(self.iaps)

        iap_sum = reduce(
            lambda result, x: result + x['normalised'],
            self.iaps,
            np.zeros(len(self.iaps[0]['normalised']))
        )

        return iap_sum * (xn / len(self.iaps))

    def __get_window_at_index(self, i):
        return [iap['normalised'][i] for iap in self.iaps]

    def __split_groups(self, threshold=0.1):
        mean_profile = self.get_mean_profile()
        N = int(round(len(mean_profile) * threshold))

        sorted_means = sorted(
            enumerate(mean_profile),
            lambda x, y: cmp(x[1], y[1])
        )

        noise = map(
            lambda x: (x[0], self.__get_window_at_index(x[0])),
            sorted_means[:N]
        )

        relevant = map(
            lambda x: (x[0], self.__get_window_at_index(x[0])),
            sorted_means
        )

        return noise, relevant

    def relevant_affective_profile(self, threshold=0.1, crit_value=0.05):
        noise, relevant = self.__split_groups(threshold)
        relevant_indices, p_values = [], []

        for relevant_window in relevant:
            p_val_sum = 0

            for noise_window in noise:
                try:
                    _, pval = mannwhitneyu(relevant_window[1], noise_window[1])
                    p_val_sum += pval
                except ValueError:
                    p_val_sum += 1

            p_val_avg = p_val_sum / len(relevant_window[1])

            if p_val_avg < crit_value:
                relevant_indices.append(relevant_window[0])

            p_values.append((relevant_window[0], p_val_avg))

        sorted_p_values = sorted(
            p_values,
            lambda x, y: cmp(x[0], y[0])
        )

        return sorted(relevant_indices), map(lambda x: x[1], sorted_p_values)


def contiguous_segments(arr):
    result = []

    if len(arr) == 0:
        return result

    start = arr[0]
    for i in xrange(len(arr) - 1):
        if arr[i+1] != arr[i] + 1:
            result.append((start, arr[i]))
            start = arr[i+1]

    result.append((start, arr[-1]))
    return result


def draw_relevant_profile(relevant, mean):
    for pair in relevant:
        y = max(mean[pair[0]:pair[1] + 1]) + 0.02
        pl.plot(list(pair), [y, y], lw=2, color="r")


def bar_no_labels(data):
    bars = pl.bar(range(len(data)), data)
    maxval = float(max(data))

    def map_color(x):
        return -(250/360.0) * x/maxval + (250/360.0)

    for i, _ in enumerate(bars):
        hue = map_color(data[i])
        bars[i].set_color(hls_to_rgb(hue, 0.5, 1))

    pl.xlim(0, len(data))


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--sample_len", type=int,
                        help="sample length of subsampled file")
    parser.add_argument("-i", "--ids", type=str,
                        help="sensor ids")
    parser.add_argument("-t", "--threshold", type=float,
                        help="signal/noise split (defaults to 0.1)")
    parser.add_argument("sensorfile")
    parser.add_argument("iaifile")
    parser.add_argument("outpath")

    args = parser.parse_args()

    threshold, ids, sample_len = 0.1, None, 30

    if args.threshold:
        threshold = float(args.threshold)
    if args.sample_len:
        sample_len = int(args.sample_len)
    if args.ids:
        ids = args.ids.split(',')

    return (
        args.sensorfile, args.iaifile, args.outpath,
        threshold, ids, sample_len
    )

if __name__ == "__main__":
    filename, iaifile, outpath, threshold, ids, sample_len = parse_args()

    normalised = read_csv(filename, index_col=0)
    iais = read_csv(iaifile, index_col=0)
    profiles = []

    if ids is None:
        ids = normalised.columns

    print "Loading datasets..."
    sys.stdout.flush()

    for sensor in ids:
        profiles.append({
            "normalised": np.array(normalised[sensor]),
            "iai": iais[sensor][0]
        })

    print "Computing individual distribution..."
    sys.stdout.flush()

    mean_profile = MeanAffectiveProfile(profiles)
    iap_distribution = mean_profile.get_iap_distribution()

    fig = pl.figure(figsize=(20, 8))
    pl.grid(False)

    pl.boxplot(iap_distribution, sym="r+")

    ticks = range(1, len(iap_distribution) + 1, 150 / (sample_len/2))
    labels = map(lambda n: (n - 1) * (sample_len / 2) / 60.0, ticks)
    pl.xticks(ticks, labels, rotation=45, fontweight='bold', fontsize=12)
    pl.yticks(fontweight="bold", fontsize=12)

    fig.savefig(outpath + "/figure_1.png")
    pl.close(fig)

    print "Computing relevant profile..."
    sys.stdout.flush()

    fig = pl.figure(figsize=(20, 9))
    pl.grid(False)
    pl.yscale('log')

    relevant, p_values = mean_profile.relevant_affective_profile(threshold)

    pl.plot(p_values)
    pl.plot([0, len(p_values)], [0.05, 0.05], lw=1, color="r")
    pl.xlim(0, len(p_values))

    ticks = range(0, len(iap_distribution), 150 / (sample_len/2))
    labels = map(lambda n: n * (sample_len / 2) / 60.0, ticks)
    pl.xticks(ticks, labels, rotation=45, fontweight='bold', fontsize=12)
    pl.yticks(fontweight="bold", fontsize=12)

    fig.savefig(outpath + "/figure_2.png")
    pl.close(fig)

    print "Generating and rendering mean profile..."
    sys.stdout.flush()

    fig = pl.figure(figsize=(20, 8))
    pl.grid(False)

    mean_data = mean_profile.get_mean_profile()

    bar_no_labels(mean_data)
    draw_relevant_profile(
        contiguous_segments(relevant),
        mean_data
    )

    index = normalised.index
    min_length = min(map(len, [index, mean_data, p_values]))

    relevant = ["TRUE" if p < 0.05 else "" for p in p_values[:min_length]]

    mean_data_df = DataFrame(
        index=index[:min_length],
        data={
            "0": mean_data[:min_length],
            "p_value": p_values[:min_length],
            "relevant": relevant
        }
    )

    mean_data_df.to_csv(outpath + "/mean_affective_profile.csv")

    relevant_df = DataFrame(data=relevant)
    relevant_df.to_csv(outpath + "/relevant_profile.csv")

    ticks = range(0, len(iap_distribution), 150 / (sample_len/2))
    labels = map(lambda n: n * (sample_len / 2) / 60.0, ticks)
    pl.xticks(ticks, labels, rotation=45, fontweight='bold', fontsize=12)
    pl.yticks(fontweight="bold", fontsize=12)

    fig.savefig(outpath + "/figure_3.png")
    pl.close(fig)
