import numpy as np
import pandas as pd
import re
import sys

from csv import Sniffer
from scipy.signal import medfilt


def already_cleaned(infile):
    with open(infile) as f:
        sample = f.read(2048)

        rows = [row for row in sample.split("\n") if not row.startswith("#")]
        cols = rows[0].split(",")

        if cols[0] == "":
            return True

        return False


def has_header(infile):
    sniffer = Sniffer()

    with open(infile) as f:
        sample = f.read(2048).replace("\r", "")
        rows = [row for row in sample.split("\n") if not row.startswith("#")]
        sample = "\n".join(rows)

        return sniffer.has_header(sample)


def convert_to_seconds(sensor, colnames={}):
    means = sensor.groupby(sensor.index)[colnames['gsr']].mean()
    return pd.DataFrame(means)


def scale_data(values, lowerbound=0, upperbound=1000):
    minval = min(values)

    if minval < lowerbound:
        values = values + abs(minval)

    maxval = max(values)

    if maxval > upperbound:
        values = values / maxval * 1000

    return values


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print "USAGE:", sys.argv[0], "input_file output_file"
        sys.exit(1)

    infile, outfile = sys.argv[1:]

    if already_cleaned(infile):
        print "File already processed. Writing to output unchanged"

        from shutil import copyfile
        copyfile(infile, outfile)

        sys.exit(0)

    if has_header(infile):
        data = pd.read_csv(infile, index_col=0, comment='#')
    else:
        print "Warning: No headers detected, setting headers manually"
        data = pd.read_csv(infile, index_col=0, comment='#', header=None)

        if len(data.columns) == 8:
            data.columns = ["", "", "nodeid", "gsr", "", "", "", ""]
        elif len(data.columns) == 4:
            data.columns = ["id", "", "", "value"]

    data.index = map(int, data.index)
    columns = data.columns
    colnames = {"nodeid": "nodeid", "gsr": "gsr"}

    if 'nodeid' not in columns:
        colnames['nodeid'] = 'id'
        colnames['gsr'] = 'value'

    sensor_ids = data[colnames['nodeid']].dropna().unique()
    timestamps = data.index.unique()
    cleaned = {}

    for sensor_id in sensor_ids:
        sensor = data[data[colnames['nodeid']] == sensor_id]
        received = len(sensor[~np.isnan(sensor[colnames['gsr']])])

        print "== PROCESSING DATA FOR SENSOR", sensor_id, "=="
        print "Received values:", received

        sensor = convert_to_seconds(sensor, colnames)
        idxdiff = np.diff(list(sensor.index))

        if max(idxdiff) > 1:
            print "WARNING: Discontinuity in timestamps detected"

        sensor_interpolated = pd.Series(
            data=sensor[colnames['gsr']], index=timestamps
        ).interpolate()

        scaled = scale_data(sensor_interpolated)
        cleaned[int(sensor_id)] = scaled

        print "OK"
        print

    pd.DataFrame(
        data=cleaned,
        index=timestamps
    ).to_csv(outfile)
